/*
** Module   : SelectionListBody.java
** Abstract : selection-list implementation
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 EVL 20050203   @19739 Creation based on CHARVA JList class.
**                           Necessary modification to get compiled
** 002 EVL 20050429   @20944 Adding label creation for the widget
** 003 EVL 20050503   @21073 Changing constructor signature to reflect
**                           the possible options for creation widget
** 004 EVL 20050503   @21074 Adding the implementation of the item
**                           manipulation methods
** 005 EVL 20050522   @21258 Change the name of the VALIDATE() method to
**                           validateThis() from runValidation() to unify
**                           the approach with the other containers
** 006 EVL 20050602   @21393 Changing the inner variable names to sync
**                           with the rest of the project
** 007 EVL 20050602   @21394 Adding the Progress specific methods to work
**                           with Progress SelectionList attributes
** 008 EVL 20050604   @21407 Fixes for javadoc comments syntax issues
** 009 EVL 20050607   @21435 Syntax for comments checking
** 010 EVL 20050607   @21438 Removing the visible member. It is completely
**                           duplicated in CHARVA's Component superclass
**                           so we do not need to double this work. Also
**                           the methods to work with the HIDDEN attribute
**                           are modified to reflect this change
** 011 EVL 20050723   @21805 Properly handling of the change the
**                           LIST_ITEMS attribute. Default delimiter is 
**                           comma.
** 012 EVL 20050822   @22199 Using the P2JFrame/Container interfaces to
**                           detect the class.
** 013 EVL 20050826   @22273 Setting up instance to a wrapped ScrollPane.
** 014 SIY 20050831   @22513 Refactoring (see @@22394-22409).
** 015 SIY 20050907   @22648 Fixed use of @inheritDoc tag in javadoc.
** 016 SIY 20050911   @22671 Updated initialisation.
** 017 EVL 20050912   @22677 Adding Validateable interface to this widget.
** 018 EVL 20050913   @22697 Adding help string handling when gaining 
**                           focus.
** 019 EVL 20050916   @22739 Returned value class is changed to be the
**                           BaseDataType.
** 020 EVL 20050920   @22761 Adding the implementation of the new method
**                           isModified() for the Validateable interface.
** 021 EVL 20050921   @22767 Adding method to be able to undo the widget's
**                           value.
** 022 SIY 20050920   @22849 Added support for SENSITIVE and SELECTED
**                           attributes. Switched to ColorTable colors.
** 023 SIY 20051129   @23563 Moved item list member to config.
** 024 GES 20060114   @24210 Implemented DataContainer interface and the
**                           core methods.
** 025 SIY 20060131   @24211 Use ThinClient special method instead of
**                           direct calls to Window.message().
** 026 SIY 20060209   @24455 Class renamed to SelectionListBody, many 
**                           other changes.
** 027 SIY 20060210   @24520 Implemented multiple selection case of 
**                           setValue().
** 028 SIY 20060222   @24643 Added necessary methods to support changes in
**                           DataContainer interface.
** 029 NVS 20060227   @24871 Key actions KA_* are used instead of virtual
**                           keys VK_* for all editing operations.
** 030 NVS 20060313   @25071 Fixed drawing of uninitialized selection list
** 031 SIY 20060515   @26158 Encapsulated access to help field.
** 032 GES 20060803   @28415 Major cleanup and refactoring.
** 033 NVS 20061204   @31560 Switched to using a private widgetFocused
**                           flag instead of config.selected
** 034 NVS 20061220   @31724 getValue() method is rewritten to return the
**                           value of the selected item, not the label.
** 035 NVS 20061221   @31732 Fixed drawing of list items (were wider by 2
**                           columns). Added removal of the highlight on
**                           GO event.
** 036 NVS 20070426   @33256 Enabled state is checked instead of SENSITIVE
**                           attribute.
** 037 SIY 20080320   @37605 Fixed handling of the SCREEN-VALUE and data
**                           value.
** 038 SIY 20080411   @37939 Fixed drawing. Cleanups and refactoring.
** 039 SIY 20080414   @37971 Fixed minor regression caused by #38. 
**                           Refactored and cleaned up draw() method.
** 040 NVS 20080625   @38927 Fixed drawing when disabled.
** 041 SIY 20100619          Switching to own implementation of character UI.
** 042 SIY 20100804          Changes in imports, some refactorings.
** 043 SIY 20101122          Switching from KeyEvent to KeyInput.
** 044 GES 20110420          Fixed querying the value in multiple selection
**                           case.
** 045 SIY 20110112          Changes related to new color handling scheme.
** 046 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 047 SIY 20111202          Refactored and moved from CHUI package.
** 048 SVL 20120821          Changed c'tor signature in order to match the one
**                           used by ChuiWidgetFactory. Added widget()
**                           function.
** 049 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 050 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 051 IAS  20150204         Fix behavior to be compatible with 4GL
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * A component that allows the user to select one or more objects from a list.
 * <p>
 * This class does not provide scrolling capability. Instances of this class
 * are normally inserted into a <code>ScrollPane</code> to provide scrolling.
 */
public abstract class SelectionListBody<O extends OutputManager<?>>
extends DefaultList<String, O>
implements FocusListener,
           SelectionListener,
           DataContainer
{
   /** Component attribute container. */
   protected SelectionListConfig config = null;
   
   /** Widget in focus flag */
   protected boolean widgetFocused = false;

   /** Default initial value. */
   private BaseDataType initValue = new character();

   /** Track user interaction with widget. */
   private boolean selected = false;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SelectionListBody(WidgetId id, SelectionListConfig cfg)
   {
      super(new ListModel<String>());
      config = (SelectionListConfig) cfg;

      addFocusListener(this);
      addSelectionListener(this);
   }

   /**
    * Requesting the focus to the <code>SelectionListBody</code>.
    */
   @Override
   public void requestFocus()
   {
      if (!isEnabled())
         return;
      super.requestFocus();
   }
   
   /**
    * Draws this component. Overrides draw() in JList.
    */
   @Override
   public abstract void draw();

   /**
    * Get item text at specified index.
    * 
    * @param   i
    *          The item index.
    *          
    * @return  The item text.
    */
   protected String getItemText(int i)
   {
      return model().get(i).toString();
   }
   
   /**
    * Key events handler for the SelectionListBody widget. The special processing
    * for SelectionListBody is the SPACE key is the same as ENTER key. Also by
    * moving the list cursor the user changes selected item for single
    * selection mode list.
    * 
    * @param   keyEvent
    *          The key event to be handled.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      // First call all KeyListener objects that may have been registered
      // for this component.
      int key = keyEvent.actionCode();
      
      // watch GO event to remove the highlight
      if (key == Keyboard.KA_GO)
      {
         widgetFocused = false;
         draw();
      }
      else if (key == Keyboard.SE_ENTRY)
      {
         // force setting of SCREEN-VALUE on entry
         config.setModified(true);
      }

      if (selectionMode() == ListSelectionModel.SINGLE_SELECTION
          && keyEvent.keyCode() == Key.VK_ENTER)
         return;

      super.processKeyEvent(keyEvent);
   }
   
   /**
    * Process key event.
    * 
    * @param key
    *          key to process
    */
   @Override
   public void processKey(int key)
   {
      switch (key)
      {
         case Key.VK_CURSOR_DOWN :
            changeRow(1);
            break;

         case Key.VK_CURSOR_UP :
            changeRow(-1);
            break;

         case Key.VK_PAGE_DOWN :
         case Key.VK_END :
         case Key.VK_HOME :
            break;
         case Key.VK_PAGE_UP :
            if (currentRow() == 0)
            {
               changeRow(0);
            }
            break;
         case 32:
            selectItem();
            EventManager.postEvent(new KeyInput(this,
                  EventType.KEY_PRESSED,
                  Keyboard.SE_VALUE_CHANGED,
                  true));
            break;
         case Key.VK_ENTER :
            EventManager.postEvent(new KeyInput(this,
                  EventType.KEY_PRESSED,
                  Keyboard.KA_RETURN,
                  true));
            break;
      }
   }

   /**
    * Process change selected row.
    * 
    * @param delta
    *          offset to next row
    */
   private void changeRow(int delta) {
      if( delta != 0) {
         scroll(delta);
      }
      if (selectionMode() == ListSelectionModel.SINGLE_SELECTION && !selected(currentRow()))
      {
         selectItem();
         EventManager.postEvent(new KeyInput(this,
               EventType.KEY_PRESSED,
               Keyboard.SE_VALUE_CHANGED,
               true));
      }
   }
   /**
    * Returns the character value of the entry with specified index from the
    * SelectionListBody.
    * 
    * @param   index
    *          The index of the entry which value should be returned.
    * 
    * @return  The string value of the entry with given index.
    */
   public String entry(int index)
   {
      return model().get(index);
   }
   
   /**
    * Checks the selection state of the entry with specified index in
    * SelectionListBody.
    * 
    * @param   index
    *          The index of the entry to be checked inside the list.
    * 
    * @return  <code>true</code> if operation was successful and
    *          <code>false</code> otherwise.
    */
   public boolean isSelected(int index)
   {
      return selected(index);
   }

   /**
    * Scrolls the list to have the specified item at the top of the list.
    * 
    * @param   item
    *          The string entry which must be visible after scrolling the
    *          list.
    * 
    * @return  <code>true</code> if operation was successful and
    *          <code>false</code> otherwise.
    */
   public boolean scrollToItem(String item)
   {
      scrollToItem(model().indexOf(item));
      return true;
   }
   
   /**
    * Scrolls the list to have the specified item at the top of the list.
    * 
    * @param   index
    *          The index of the entry which must be visible after scrolling
    *          the list.
    * 
    * @return  <code>true</code> if operation was successful and
    *          <code>false</code> otherwise.
    */
   public boolean scrollToItem(int index)
   {
      ensureRowVisible(index);
      return true;
   }
   
   /**
    * Getting the currently selected item.
    * Since the list is made of {label, value} pairs, this method returns
    * the value.
    *
    * @return  The BaseDataType object representing the current value of
    *          the selection-list.
    */
   @Override
   public BaseDataType getValue()
   {
      if (selected )
         return getScreenValue();
      
      return initValue;
   }

   /**
    * Sets the BaseDataType instance representing the current value.
    * 
    * @param   value
    *          The value representing the current widget value.
    */
   @Override
   public void setValue(BaseDataType value)
   {  
      selected  = false;
      initValue = new character();
      
      boolean doRepaint = false;
      
      if (value.isUnknown())
         return;
      
      clear();
      
      if (value instanceof character)
      {
         String itemList = ((character) value).getValue();
         
         char[] delim = new char[1];
         delim[0] = config.delimiter;
         String d = new String(delim);
         
         String[] elem = itemList.split(d);
         
         for (String element : elem)
         {
            BaseDataType val = 
               DisplayFormat.instanceOfType(config.dataType, element);
            
            if (setSingleValue(val))
               doRepaint = true;
         }
      }
      else
         doRepaint = setSingleValue(value);

      if (doRepaint)
      {
         repaint();
         initValue = value; 
      }
      else
         if (CompareOps.isEqual(value, "").booleanValue())
            initValue = value;
   }
   
   /**
    * Set single value in the list as selected.
    * 
    * @param   value
    *          Value to set.
    * @return  <code>true</code> if value is set and repaint is required.
    */
   @SuppressWarnings("unchecked")
   private boolean setSingleValue(BaseDataType value)
   {
      ControlSetItem[] items = config.items;
      boolean doRepaint = false;
      
      for (int i = 0; i < items.length; i++)
      {
         if (items[i].getValue().compareTo(value) == 0)
         {
            select(i,i);
            
            if (first() >= 0)
               ensureRowVisible(first());
            
            doRepaint = true;
         }
      }
      return doRepaint;
   }

   /**
    * The implementation of the ListSelectionListener interface. Marks the
    * widget as modified.
    */
   @Override
   public void changed()
   {
      config.setModified(true);
      selected = true;
   }
   
   /**
    * Getting the modification state of the widget.
    * 
    * @return  <code>true</code> if the widget value was modified,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean isModified()
   {
      return config.modified;
   }
   
   /**
    * This is called when the selection-list takes the input focus. When the
    * help string is defined for the selection-list it should be displayed at
    * the window status line.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusGained(FocusEvent focusEvent)
   {
      if (focusEvent == null)
         return;

      widgetFocused = true;
      
      // Draw the help string for focused button
      ThinClient.setHelp(config.help);
   }

   /**
    * This is called when the selection-list looses the input focus.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusLost(FocusEvent focusEvent)
   {
      if (focusEvent == null)
         return;
      
      widgetFocused = false;
   }

   /**
    * Update list of items from configuration.
    */
   public void refreshItems()
   {
      model().clear();
      
      ControlSetItem[] items = config.items;
      
      for (ControlSetItem element : items)
         model().add(element.getLabel().getValue());
   }

   /**
    * Get state of the HORIZONTAL-SCROLL-BAR attribute.
    * 
    * @return  <code>true</code> if scroll bar is enabled.
    */
   public boolean isHorizontalScrollBar()
   {
      return config.scrollbarHorizontal;
   }
 
   /**
    * Get state of the VERTICAL-SCROLL-BAR attribute.
    * 
    * @return  <code>true</code> if scroll bar is enabled.
    */
   public boolean isVerticalScrollBar()
   {
      return config.scrollbarVertical;
   }

   /**
    * Get the text representation of the currently selected value. This will be
    * rendered as the labels that are displayed (which are the same as the
    * value in LIST-ITEMS mode, but are different from the value in 
    * LIST-ITEM-PAIRS mode).
    * <p>
    * This will be a single text value if this is a single-select list and
    * will be a delimited list if this is multiple-select.
    * 
    * @return   The screen representation of the value or empty string if there
    *           is no selection.
    */
   @Override
   public String getText()
   {
      String result = null;

      for (int i = 0; i < model().size(); i++)
      {
         if (isSelected(i))
         {
            if (result == null)
               result = entry(i);
            else
               result += Character.toString(config.delimiter) + entry(i);
         }
      }

      return (result == null) ? "" : result;
   }
   
   /**
    * Get widget screen value (suitable for assigning back to any associated
    * variable or field). This will be rendered as the values (which are the
    * same as the labels in LIST-ITEMS mode, but are different from the labels 
    * in LIST-ITEM-PAIRS mode).
    * <p>
    * This will be a single text value if this is a single-select list and
    * will be a delimited list if this is multiple-select.
    * 
    * @return  Current widget screen value.
    */
   public BaseDataType getScreenValue()
   {
      ControlSetItem[] items = config.items;
      
      int[]  sel   = selection();
      String delim = Character.toString(config.delimiter);
      
      StringBuilder sb = new StringBuilder();
      
      for (int i = 0; i < sel.length; i++)
      {
         if (sel[i] >= 0 && sel[i] < items.length)
         {
            character val = (character) items[sel[i]].getValue();
            
            if (i > 0)
               sb.append(delim);
            
            sb.append(val.toStringMessage());
         }
      }
      
      return sb.length() == 0 ? new character() : new character(sb.toString());
   }

   /**
    * Provide access to the attributes stored in the widget.
    *
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public SelectionListConfig config()
   {
      return config;
   }
}
